/* GLSL_toon: A textured toon shader with sobel edge detection
   Copyright (C) 2009 Angelo "Encelo" Theodorou

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <iostream>
#include <cmath>

#include <SDL/SDL.h>
#include <GL/gl.h>
#include <GL/glext.h>
#include <GL/glu.h>

#include "Shader.h"
#include "ShaderProgram.h"
#include "FrameTimer.h"
#include "VertexBuffer.h"
#include "Mesh.h"
#include "Light.h"
#include "Texture.h"
#include "FrameBuffer.h"

using std::cout;
using std::endl;

int main(int argc, char **argv)
{
	SDL_Surface *screen;
	SDL_Event event;
	const SDL_VideoInfo* info = NULL;
	int width = 640;
	int height = 480;
	int bpp = 0;
	int flags = 0;

	bool quit = false;
	bool rotate = false, lmove = true;
	GLfloat rotcam = 0.0f, lrot = 0.0f;
	bool with_edge = true, textured = true;

	const float default_thickness = 1.5f;
	const float thickness_step = 0.1f;
	float thickness = default_thickness;
	const float default_threshold = 0.3;
	const float threshold_step = 0.05;
	float threshold = default_threshold;

	FrameTimer t(5);
	Light light;

// ----- SDL init ---------------
	if (SDL_Init(SDL_INIT_VIDEO) < 0) {
		fprintf(stderr, "Video initialization failed: %s\n", SDL_GetError());
		exit(-1);
	}

	atexit(SDL_Quit);
	
	info = SDL_GetVideoInfo();
	bpp = info->vfmt->BitsPerPixel;

	
// ----- OpenGL attribute setting via SDL ---------------
	SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	SDL_GL_SetAttribute(SDL_GL_SWAP_CONTROL, 0);
	
	flags = SDL_OPENGL;

	
// ----- Setting up the screen surface ---------------
	if ((screen = SDL_SetVideoMode(width, height, bpp, flags)) == 0) {
		fprintf(stderr, "Video mode set failed: %s\n", SDL_GetError());
        	exit(-1);
	}

	SDL_WM_SetCaption("GLSL_toon by Encelo", NULL);


// ----- OpenGL init ---------------
	glClearColor(0.25f, 0.25f, 0.25f, 0.0f);
	glEnable(GL_DEPTH_TEST);

	ShaderProgram toon("shaders/toon.vert", "shaders/toon.frag");
	ShaderProgram edge("shaders/sobel.vert", "shaders/sobel.frag");
	edge.Uniform("Tex0", 0);
	edge.Uniform("Tex1", 1);
	edge.Uniform("Width", width);
	edge.Uniform("Height", height);

	FrameBuffer fbo(width, height);
	fbo.AttachRender(GL_DEPTH_COMPONENT24);
    fbo.AttachTexture(GL_RGBA8);
	fbo.AttachTexture(GL_RGBA8);
	FrameBuffer::Unbind();

	Texture tex("textures/head_toon.png");
	Mesh head("models/uv_head.ply");

	VertexBuffer quad;
	GLfloat qverts[] = {-1,-1, 0,  1, -1, 0,  1, 1, 0,  -1, 1, 0};
	GLfloat qcoords[] = {0, 0,  1, 0,  1, 1,  0, 1};
	GLuint qind[] = {0, 1, 2,  0, 2, 3};
	quad.SetVertices(4, qverts);
	quad.SetTexCoords(4, qcoords);
	quad.SetIndices(6, qind);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0f, float(width)/float(height), 0.1f, 10.0f);
	t.Reset();


// ----- Event cycle ---------------
	while (!quit) {
		while (SDL_PollEvent(&event)) {
			switch (event.type) { 
			case SDL_QUIT:
				quit = true;
				break;
			case SDL_KEYDOWN:
				switch (event.key.keysym.sym) {
				case SDLK_ESCAPE:
				case SDLK_q:
					quit = true;
					break;
				case SDLK_F1:
					SDL_WM_ToggleFullScreen(screen);
					break;
				case SDLK_l:
					lmove = !lmove;
					break;
				case SDLK_r:
					rotate = !rotate;
					break;
				case SDLK_e:
					with_edge = !with_edge;
					break;
				case SDLK_t:
					textured = !textured;
					break;
				case SDLK_KP0:
					with_edge = true;
					thickness = default_thickness;
					threshold = default_threshold;
					rotcam = lrot = 0.0f;
					break;
				case SDLK_KP7:
					thickness -= thickness_step;
					break;
				case SDLK_KP8:
					thickness = default_thickness;
					break;
				case SDLK_KP9:
					thickness += thickness_step;
					break;
				case SDLK_KP4:
					threshold += threshold_step;
					break;
				case SDLK_KP5:
					threshold = default_threshold;
					break;
				case SDLK_KP6:
					threshold -= threshold_step;
					break;
				default:
					break;
				}
			}
		}

		
// ----- Blitting on the screen ---------------
		t.AddFrame();
		if (rotate) {
			rotcam += 0.001f * t.GetInterval();
		}
		if (lmove) {
			lrot += 0.002f * t.GetInterval();
		}

		// First Pass: scene toon rendering
		if (with_edge)
			fbo.BindIn();
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		toon.Use();
		toon.Uniform("Textured", textured);

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		light.Position(2 * sinf(lrot), 0.0f, 0.0f);
		glTranslatef(0.0f, 0.0f, -3.0f);
		glRotatef(15.0f, 1.0f, 0.0f, 0.0f);
		glRotatef(30 * sinf(rotcam), 0.0f, 1.0f, 0.0f);

		glActiveTexture(GL_TEXTURE0);
		tex.Bind();
		head.Draw();

		if (with_edge) {

			glLoadIdentity();
			glMatrixMode(GL_PROJECTION);
			glPushMatrix();
			glLoadIdentity();

			// Second pass: edge detection and drawing
			FrameBuffer::Unbind();
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
			glActiveTexture(GL_TEXTURE0);
			fbo.BindTex(0);
			glActiveTexture(GL_TEXTURE1);
			fbo.BindTex(1);
			edge.Use();
			edge.Uniform("Thickness", thickness);
			edge.Uniform("Threshold", threshold);
			quad.DrawElements();

			glPopMatrix();
			glMatrixMode(GL_MODELVIEW);
		}

		SDL_GL_SwapBuffers();		
	}

// ----- Quitting ---------------
	SDL_Quit();
	return 0;	
}
