#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <stdexcept>
#include <cmath>

#include "Mesh.h"
#include "VertexBuffer.h"

using std::ios;
using std::ifstream;
using std::string;
using std::istringstream;
using std::invalid_argument;
using std::domain_error;
using std::cout;
using std::endl;

#include "Timer.h" // XXX: for debug purposes only
Timer t;


Mesh::Mesh(const string& filename) throw(invalid_argument) : _nvert(0), _ntri(0),
	_vertices(NULL), _normals(NULL), _texcoords(NULL), _tangents(NULL), _indices(NULL)
{
	string::size_type loc = filename.find(".ply", 0);

	if (loc != filename.length()) {
		cout << "Loading " << filename << endl;
		t.Start();
		LoadPLY(filename.c_str());
		t.Stop();
		cout << "PLY loading: " << t.GetInterval() << "ms" << endl;
	}
	else {
		throw invalid_argument("Mesh::Mesh - unrecognized file extension");
	}

	SetupVBO();
}


void Mesh::LoadPLY(const char *filename) throw(domain_error)
{
	int i;

	string str;
	istringstream ss;
	string::size_type pos;

	ifstream is;
	is.open(filename, ios::in);

	int nprop = 0;

	// Parsing the ply identification line
	std::getline(is, str);
	if (str != "ply") {
		throw domain_error("Mesh::LoadPly - Not a PLY file");
	}

	// Parsing the format specification line
	std::getline(is, str);
	if (str != "format ascii 1.0") {
		throw domain_error("Mesh::LoadPly - Not an ASCII PLY file");
	}


	// Parsing number of vertices
	while (is.good()) {
		std::getline(is, str);

		if ( (pos = str.find("element vertex ")) == string::npos ) {
			continue;
		}
		else {
			ss.str( str.substr(string("element vertex ").length()) );
			ss >> _nvert;
			ss.clear(istringstream::goodbit);

			cout << "Vertices: " << _nvert << endl;
			break;
		}
	}


	// Parsing number of vertex properties
	while (is.good()) {
		std::getline(is, str);

		if ( (pos = str.find("property ")) != string::npos ) {
			nprop++;
		}
		else {
			is.seekg(-is.gcount()); // Unget last line
			cout << "Properties: " << nprop << endl;
			break;
		}
	}


	// Parsing number of triangles
	while (is.good()) {
		std::getline(is, str);

		if ( (pos = str.find("element face ")) == string::npos ) {
			continue;
		}
		else {
			ss.str( str.substr(string("element face ").length()) );
			ss >> _ntri;
			ss.clear(istringstream::goodbit);

			cout << "Triangles: " << _ntri << endl;
			break;
		}

	}

	// Parsing end of the header
	while (is.good()) {
		std::getline(is, str);

		if ( (pos = str.find("end_header")) == string::npos ) {
			continue;
		}
		else {
			break;
		}
	}


	_vertices = new GLfloat[_nvert * 3];
	if (nprop >= 6) {
		_normals = new GLfloat[_nvert * 3];
	}
	if (nprop >= 8) {
		_texcoords = new GLfloat[_nvert * 2];
	}

	// Parsing vertices
	for (i = 0; i < _nvert; i++) {
		std::getline(is, str);

		ss.str(str);
		ss >> _vertices[i*3] >> _vertices[i*3 + 1] >> _vertices[i*3 + 2];
		if (nprop >= 6) {
			ss >> _normals[i*3] >> _normals[i*3 + 1] >> _normals[i*3 + 2];
		}
		if (nprop >= 8) {
			ss >> _texcoords[i*2] >> _texcoords[i*2 + 1];
		}
		ss.clear(istringstream::goodbit);

//		cout << "Vertices: " << _vertices[i*3] << " " << _vertices[i*3 + 1] << " " << _vertices[i*3 + 2] << endl;
//		cout << "Normals: " << _normals[i*3] << " " << _normals[i*3 + 1] << " " << _normals[i*3 + 2] << endl;
//		cout << "TexCoords: " << _texcoords[i*2] << " " << _texcoords[i*2 + 1] << endl;
	}

	_indices = new unsigned int[_ntri * 3];

	// Parsing triangles
	int face_vert;
	for (i = 0; i < _ntri; i++) {
		std::getline(is, str);

		ss.str(str);
		ss >> face_vert;
		if (face_vert != 3) {
			throw domain_error("Mesh::LoadPly - Not a triangular face");
		}
		ss >> _indices[i*3] >> _indices[i*3 + 1] >> _indices[i*3 + 2];
		ss.clear(istringstream::goodbit);

//		cout << "Indices: " << _indices[i*3] << " " << _indices[i*3 + 1] << " " << _indices[i*3 + 2] << endl;
	}

	is.close();
}


void Mesh::SetupVBO()
{
	if (_vertices) {
		_vbo.SetVertices(_nvert, _vertices);
		delete[] _vertices;
		_vertices = NULL;
	}
	
	if (_normals) {
		_vbo.SetNormals(_nvert, _normals);
		delete[] _normals;
		_normals = NULL;
	}

	if (_texcoords) {
		_vbo.SetTexCoords(_nvert, _texcoords);
		delete[] _texcoords;
		_texcoords = NULL;
	}

	if (_indices) {
		 _vbo.SetIndices(_ntri*3, _indices);
		 delete[] _indices;
		 _indices = NULL;
	}

}


void Mesh::Draw()
{
	_vbo.DrawElements();
}
